import os
import torch
import torchvision.datasets as dsets
import re

from datasets.templates import get_templates
from datasets.common import get_target_transform


def pretify_classname(classname):
    l = re.findall(r'[A-Z](?:[a-z]+|[A-Z]*(?=[A-Z]|$))', classname)
    l = [i.lower() for i in l]
    out = ' '.join(l)
    if out.endswith('al'):
        return out + ' area'
    return out

class EuroSATBase:
    def __init__(self,
                 preprocess,
                 test_split,
                 location='~/datasets',
                 batch_size=32,
                 num_workers=16,
                 image_text=False):
        # Data loading code
        traindir = os.path.join(location, 'EuroSAT_splits', 'train')
        testdir = os.path.join(location, 'EuroSAT_splits', test_split)

        if (image_text):
            self.train_dataset = dsets.ImageFolder(traindir, transform=preprocess)
            idx_to_class = dict((v, k)
                                for k, v in self.train_dataset.class_to_idx.items())
            self.classnames = [idx_to_class[i].replace('_', ' ') for i in range(len(idx_to_class))]
            self.classnames = [pretify_classname(c) for c in self.classnames]
            ours_to_open_ai = {
                'annual crop': 'annual crop land',
                'forest': 'forest',
                'herbaceous vegetation': 'brushland or shrubland',
                'highway': 'highway or road',
                'industrial area': 'industrial buildings or commercial buildings',
                'pasture': 'pasture land',
                'permanent crop': 'permanent crop land',
                'residential area': 'residential buildings or homes or apartments',
                'river': 'river',
                'sea lake': 'lake or sea',
            }
            for i in range(len(self.classnames)):
                self.classnames[i] = ours_to_open_ai[self.classnames[i]]
            templates = get_templates('EuroSAT')
            target_transform = get_target_transform(templates, self.classnames)

        self.train_dataset = dsets.ImageFolder(traindir, transform=preprocess,
                                               target_transform=target_transform if image_text else None)
        self.train_loader = torch.utils.data.DataLoader(
            self.train_dataset,
            shuffle=True,
            batch_size=batch_size,
            num_workers=num_workers,
        )

        self.test_dataset = dsets.ImageFolder(testdir, transform=preprocess,
                                              target_transform=target_transform if image_text else None)
        self.test_loader = torch.utils.data.DataLoader(
            self.test_dataset,
            batch_size=batch_size,
            num_workers=num_workers
        )
        idx_to_class = dict((v, k)
                            for k, v in self.train_dataset.class_to_idx.items())
        self.classnames = [idx_to_class[i].replace('_', ' ') for i in range(len(idx_to_class))]
        self.classnames = [pretify_classname(c) for c in self.classnames]
        ours_to_open_ai = {
            'annual crop': 'annual crop land',
            'forest': 'forest',
            'herbaceous vegetation': 'brushland or shrubland',
            'highway': 'highway or road',
            'industrial area': 'industrial buildings or commercial buildings',
            'pasture': 'pasture land',
            'permanent crop': 'permanent crop land',
            'residential area': 'residential buildings or homes or apartments',
            'river': 'river',
            'sea lake': 'lake or sea',
        }
        for i in range(len(self.classnames)):
            self.classnames[i] = ours_to_open_ai[self.classnames[i]]


class EuroSAT(EuroSATBase):
    def __init__(self,
                 preprocess,
                 location='~/datasets',
                 batch_size=32,
                 num_workers=16,
                 image_text=False):
        super().__init__(preprocess, 'test', location, batch_size, num_workers, image_text=image_text)


class EuroSATVal(EuroSATBase):
    def __init__(self,
                 preprocess,
                 location='~/datasets',
                 batch_size=32,
                 num_workers=16,
                 image_text=False):
        super().__init__(preprocess, 'val', location, batch_size, num_workers, image_text=image_text)